/**
 * © 2008 by David ‘Bombe’ Roden <bombe@pterodactylus.net>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "GrowingBuffer.h"
#include "GlobalSettings.h"

GrowingBuffer::GrowingBuffer(size_t initialSize) {
	this->data = malloc(initialSize);
	this->size = initialSize;
	this->limit = 0;
	this->position = 0;
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] created new GrowingBuffer with a size of %d.\n", __FILE__, __LINE__, initialSize);
}

GrowingBuffer::~GrowingBuffer() {
	if (this->data) {
		free(this->data);
	}
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] destroyed GrowingBuffer.\n", __FILE__, __LINE__);
}

size_t GrowingBuffer::getPosition() {
	return position;
}

size_t GrowingBuffer::getLimit() {
	return limit;
}

size_t GrowingBuffer::getSize() {
	return size;
}

void GrowingBuffer::seek(size_t position) {
	this->position = (position > limit) ? limit : position;
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] setting position to %d.\n", __FILE__, __LINE__, position);
}

void* GrowingBuffer::read(void* buffer, size_t length) {
	size_t bytesToCopy = (length > (limit - position)) ? (limit - position) : length;
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] copying %d bytes (%d requested) into buffer from position %d.\n", __FILE__, __LINE__, bytesToCopy, length, position);
	memcpy(buffer, (char*) data + position, bytesToCopy);
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] advancing position from %d to %d.\n", __FILE__, __LINE__, position, position + bytesToCopy);
	position += bytesToCopy;
	return buffer;
}

void GrowingBuffer::write(const void* buffer, size_t length) {
	if (length > (size - limit)) {
		int newSize = size;
		do {
			newSize *= 2;
		} while (length > (newSize - limit));
		GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] resizing buffer from %d to %d bytes to fit in %d bytes.\n", __FILE__, __LINE__, size, newSize, length);
		void* newData = malloc(newSize);
		GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] copying %d bytes into new buffer.\n", __FILE__, __LINE__, limit);
		memcpy(newData, data, limit);
		free(data);
		data = newData;
	}
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] writing %d bytes to buffer at position %d.\n", __FILE__, __LINE__, length, limit);
	memcpy((char*) data + limit, buffer, length);
	limit += length;
}

void GrowingBuffer::cut() {
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] discarding %d bytes.\n", __FILE__, __LINE__, position);
	memcpy(data, (char*) data + position, position);
	limit -= position;
	position = 0;
}

void GrowingBuffer::truncate() {
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] truncating %d bytes at position %d.\n", __FILE__, __LINE__, (limit - position), position);
	limit = position;
}

void GrowingBuffer::clear() {
	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] clearing buffer, discarding %d bytes.\n", __FILE__, __LINE__, limit);
	position = 0;
	limit = 0;
}

void GrowingBuffer::resize(double factor) {
	if (factor < 1.0) {
		factor = 1.0;
	}
	if (size != (limit * factor)) {
		GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] resizing buffer from %d to %d bytes.\n", __FILE__, __LINE__, size, (size_t) (limit * factor));
		void* newData = malloc((size_t) (limit * factor));
		memcpy(newData, data, limit);
		free(data);
		data = newData;
	}
}

size_t GrowingBuffer::indexOf(char c, size_t start) {
	void* foundIndex;

	GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] searching for (char) 0x%02x from %d to %d.\n", __FILE__, __LINE__, c, position + start, position + (limit - start));
	foundIndex = memchr((char*) data + position + start, c, (limit - (position + start)));
	if (foundIndex) {
		GlobalSettings::isVerbose() && fprintf(stderr, "[%s:%d] found character at %d.\n", __FILE__, __LINE__, (char*) foundIndex - ((char*) data + position));
		return (char*) foundIndex - ((char*) data - position);
	}
	return (size_t) -1;
}

